import {desiredHeight, desiredWidth, getScreen} from "./engine/internal/screen.ts";
import {BG_INSET, BG_OUTER, FG_TEXT} from "./colors.ts";
import {D, I} from "./engine/public.ts";
import {IGame, Point, Rect, Size} from "./engine/datatypes.ts";
import {sprDrips, sprRaccoonWalking, sprStatPickup} from "./sprites.ts";
import {DrawPile} from "./drawpile.ts";
import {HuntMode, MapCell} from "./huntmode.ts";
import {ALL_STATS} from "./datatypes.ts";

class MenuCamera {
  // measured in whole screens
  position: Point;
  target: Point;

  constructor({position, target}: {position: Point, target: Point}) {
    this.position = position;
    this.target = target;
  }

  update() {
    let adjust = (x0: number, x1: number) => {
      if (Math.abs(x1 - x0) < 0.01) { return x1; }
      return (x0 * 8 + x1 * 2) / 10;
    }
    this.position = new Point(
      adjust(this.position.x, this.target.x),
      adjust(this.position.y, this.target.y),
    );
  }
}

type GameState = "Gameplay" | "Thralls";

function getScreenLocation(state: GameState): Point {
  if (state === "Gameplay") {
    return new Point(0, 0);
  }
  if (state === "Thralls") {
    return new Point(0, 1);
  }

  throw `invalid state: ${state}`
}

export class Game implements IGame {
  camera: MenuCamera;
  state: GameState;
  huntMode: HuntMode;
  gameplayDrawPile: DrawPile | null;
  frame: number;

  constructor() {
    this.camera = new MenuCamera({
      position: new Point(0, 0),
      target: new Point(0, 0),
    });
    this.state = "Gameplay";

    this.huntMode = HuntMode.generate({depth: 1});
    this.gameplayDrawPile = null;
    this.frame = 0;
  }

  update() {
    this.frame += 1;
    if (I.isKeyPressed("w")) {
      this.state = "Gameplay"
    }
    if (I.isKeyPressed("s")) {
      this.state = "Thralls"
    }

    this.camera.target = getScreenLocation(this.state);
    D.camera = new Point(
      D.size.w * this.camera.position.x,
      D.size.h * this.camera.position.y,
    )
    this.camera.update();

    // state-specific updates
    this.updateGameplay();
  }

  draw() {
    // draw screen background
    let oldCamera = D.camera;
    D.camera = new Point(0, 0);
    D.fillRect(new Point(0, 0), D.size, BG_OUTER);
    D.camera = oldCamera;

    this.drawGameplay();

    // we draw all states at once and pan between them
    // mainFont.drawText({ctx: ctx, text: "You have been given a gift.", x: 0, y: 0})
    let mouse = I.mousePosition?.offset(D.camera);
    if (mouse != null) {
      D.invertRect(mouse.offset(new Point(-1, -1)), new Size(3, 3))
    }
  }

  getPaneRegionForGameState(gameState: GameState) {
    let screen = getScreen();
    let {w, h} = screen.size;
    let overallScreenLocation = getScreenLocation(gameState);

    let bigPaneX = overallScreenLocation.x * w;
    let bigPaneY = overallScreenLocation.y * h;
    let bigPaneW = w;
    let bigPaneH = h;

    let smallPaneW = desiredWidth;
    let smallPaneH = desiredHeight;
    let smallPaneX = Math.floor(bigPaneX + (bigPaneW - smallPaneW) / 2)
    let smallPaneY = Math.floor(bigPaneY + (bigPaneH - smallPaneH) / 2)

    return {
      big: {
        position: new Point(bigPaneX, bigPaneY),
        size: new Size(bigPaneW, bigPaneH),
      },
      small: {
        position: new Point(smallPaneX, smallPaneY),
        size: new Size(smallPaneW, smallPaneH),
      }
    }
  }

  #moveCameraForGameplayDrawpile(cb: () => void) {
    let region = this.getPaneRegionForGameState("Gameplay")

    let oldCamera = D.camera;
    D.camera = D.camera.offset(region.small.position.negate());

    cb();

    D.camera = oldCamera;
  }

  updateGameplay() {
    var drawpile = new DrawPile();

    this.#moveCameraForGameplayDrawpile(() => {
      let globalOffset =
        new Point(this.huntMode.player.x * MAP_CELL_ONSCREEN_SIZE.w, this.huntMode.player.y * MAP_CELL_ONSCREEN_SIZE.h).offset(
          new Point(-192, -192)
        )

      let map = this.huntMode.cells;
      for (let y = 0; y < map.size.h; y += 1) {
        for (let x = 0; x < map.size.w; x += 1) {
          let cellOffset = new Point(x * MAP_CELL_ONSCREEN_SIZE.w, y * MAP_CELL_ONSCREEN_SIZE.h).offset(globalOffset.negate());
          let cellData = this.huntMode.cells.get(new Point(x, y))
          let belowIsBlock = true;
          if (y < map.size.h - 1) {
            let below = this.huntMode.cells.get(new Point(x, y + 1));
            belowIsBlock = !below.revealed || below.content.type == "block";
          }

          this.#drawMapCell(drawpile, cellOffset, new Point(x, y), cellData, belowIsBlock);
        }
      }
      this.#drawPlayer(drawpile, globalOffset);

      drawpile.executeOnClick();
    });

    this.gameplayDrawPile = drawpile;
  }

  drawGameplay() {
    // TODO: Draw

    this.#moveCameraForGameplayDrawpile(() => {
      this.gameplayDrawPile?.draw();

      // D.drawText("shapes", new Point(0, 0), FG_TEXT);
    });
  }

  #drawMapCell(
    drawpile: DrawPile,
    cellOffset: Point,
    mapPosition: Point,
    cellData: MapCell,
    belowIsBlock: boolean
  ) {
    const OFFSET_FLOOR = -256;
    const OFFSET_AIR = 0;
    const depth = mapPosition.y;
    const onFloor = OFFSET_FLOOR + depth;
    const inAir = OFFSET_AIR + depth;

    if (!cellData.revealed) {
      return;
    }

    let cellTopLeft = cellOffset.offset(new Size(-MAP_CELL_ONSCREEN_SIZE.w / 2, -MAP_CELL_ONSCREEN_SIZE.h / 2));
    let cellSize = MAP_CELL_ONSCREEN_SIZE;

    if (cellData.content.type == "block") {
      if (!belowIsBlock) {
        drawpile.add(inAir, () => {
          D.drawSprite(sprDrips, cellOffset.offset(new Point(0, -cellSize.h)), 1, {xScale: 3, yScale: 3})
        })
      }
      return;
    }

    // draw inset zone
    drawpile.addClickable(onFloor,
      (hover: boolean) => {
        D.fillRect(cellTopLeft, cellSize, hover ? FG_TEXT : BG_INSET)
      },
      new Rect(cellTopLeft, cellSize),
      cellData.nextMoveAccessible,
      () => {
        this.huntMode.movePlayerTo(mapPosition)
      }
    );

    if (belowIsBlock) {
      // draw the underhang
      drawpile.add(onFloor, () => {
        D.drawSprite(sprDrips, cellOffset.offset(new Point(0, cellSize.h/2)), 0, {xScale: 3, yScale: 3})
      })
    }

    if (cellData.content.type == "statPickup") {
      let content = cellData.content;
      let extraXOffset = 0; // Math.cos(this.frame / 80 + mapPosition.x + mapPosition.y) * 1;
      let extraYOffset = Math.sin(this.frame / 50 + mapPosition.x * 2+ mapPosition.y * 0.75) * 6 - 18;
      drawpile.add(inAir, () => {
        D.drawSprite(
          sprStatPickup,
          cellOffset.offset(new Point(extraXOffset, extraYOffset)),
          ALL_STATS.indexOf(content.stat),
          {
            xScale: 3,
            yScale: 3,
          }
        )
      });
    }
  }

  #drawPlayer(drawpile: DrawPile, globalOffset: Point) {
    let cellOffset = new Point(this.huntMode.player.x * MAP_CELL_ONSCREEN_SIZE.w, this.huntMode.player.y * MAP_CELL_ONSCREEN_SIZE.h).offset(globalOffset.negate())
    drawpile.add(this.huntMode.player.y, () => {
      D.drawSprite(
        sprRaccoonWalking,
        cellOffset.offset(new Point(0, 22)),
        0, {
          xScale: 3,
          yScale: 3
        }
      )
    });
  }
}

const MAP_CELL_ONSCREEN_SIZE: Size = new Size(96, 48)


export let game = new Game();