import {Grid, Point, Size} from "./engine/datatypes.ts";
import {ConceptualCell, maps} from "./maps.ts";
import {ALL_STATS, Stat} from "./datatypes.ts";

export type MapCellContent =
  {type: "statPickup", stat: Stat} |
  {type: "stairs"} |
  {type: "empty"} |
  {type: "block"}

export type MapCell = {
  content: MapCellContent,
  isValidSpawn: boolean,
  revealed: boolean,
  nextMoveAccessible: boolean,
}

export class HuntMode {
  depth: number
  cells: Grid<MapCell>
  player: Point

  constructor({depth, cells, player}: {depth: number, cells: Grid<MapCell>, player: Point }) {
    this.depth = depth;
    this.cells = cells;
    this.player = player;
  }

  // == map generator ==
  static generate({depth}: {depth: number}) {
    let mapNames: Array<string> = Object.keys(maps);
    let mapName = mapNames[Math.floor(Math.random() * mapNames.length)];
    let map = maps[mapName];

    let baseCells = map.map((ccell, _xy) => {
      return this.#generateCell(ccell);
    })

    let cells = new Grid(
      new Size(baseCells.size.w + 2, baseCells.size.h + 2), (xy) => {
        let offset = xy.offset(new Point(-1, -1));
        if (offset.x == -1 || offset.y == -1 || offset.x == baseCells.size.w || offset.y == baseCells.size.h) {
          return this.#generateBoundaryCell();
        }
        return baseCells.get(offset)
      }
    )

    let validSpawns = [];
    for (let x = 0; x < cells.size.w; x++) {
      for (let y = 0; y < cells.size.h; y++) {
        let position = new Point(x, y);
        if (cells.get(position).isValidSpawn)  {
          validSpawns.push(position);
        }
      }
    }
    let player = choose(validSpawns);
    cells.get(player).content = {type: "empty"};

    if (Math.random() < 0.75) {
      while (true) {
        let x = Math.floor(Math.random() * cells.size.w);
        let y = Math.floor(Math.random() * cells.size.h);
        let xy = new Point(x, y);

        let item = cells.get(new Point(x, y));
        if (player.equals(xy)) {
          continue;
        }
        if (item.content.type == "block") {
          continue;
        }
        item.content = {type: "stairs"}
        break;
      }
    }

    let hm = new HuntMode({depth, cells, player})
    hm.#updateVisibilityAndPossibleMoves();
    return hm;
  }

  static #generateCell(conceptual: ConceptualCell): MapCell {
    switch (conceptual) {
      case "X":
        return { content: {type: "block"}, revealed: false, isValidSpawn: false, nextMoveAccessible: false};
      case " ":
        return { content: HuntMode.#generateContent(), revealed: false, isValidSpawn: false, nextMoveAccessible: false };
      case ".":
        return { content: HuntMode.#generateContent(), revealed: false, isValidSpawn: true, nextMoveAccessible: false };
    }
  }

  static #generateBoundaryCell() {
    return this.#generateCell("X");
  }

  static #generateContent(): MapCellContent {
    // stat pickup
    let gsp = (): MapCellContent => {
      return {type: "statPickup", stat: choose(ALL_STATS)}
    };
    // TODO: Other objects?
    return choose([
      gsp, gsp, gsp, gsp
    ])();
  }

  // == update logic ==
  #updateVisibilityAndPossibleMoves() {
    for (let x = 0; x < this.cells.size.w; x++) {
      for (let y = 0; y < this.cells.size.h; y++) {
        let position = new Point(x, y);
        let data = this.cells.get(position);

        data.nextMoveAccessible = false;
        if (
          Math.abs(x - this.player.x) <= 1 &&
          Math.abs(y - this.player.y) <= 1
        ) {
          data.revealed = true;
          if (!this.player.equals(position)) {
            data.nextMoveAccessible = true;
          }
        }

      }
    }
  }

  #collectResources() {
    let present = this.cells.get(this.player);
    if (present.content.type == "statPickup") {
      present.content = {type: "empty"};
    }
  }

  movePlayerTo(newPosition: Point) {
    this.player = newPosition;
    this.#updateVisibilityAndPossibleMoves();
    this.#collectResources();
  }
}

function choose<T>(array: Array<T>): T  {
  if (array.length == 0) {
    throw `array cannot have length 0 for choose`
  }
  return array[Math.floor(Math.random() * array.length)]
}
